/***********************************************************************

AGB-DTVD v0.06 BETA by Herman Haverkort, 10 May 2022

************************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of AGB-DTVD may be available from 
http://herman.haverkort.net.


************************************************************************

Was ist AGB-DTVD?

AGB-DTVD implements three algorithms to compute Voronoi diagrams and 
Delaunay triangulations. Its purpose is to illustrate the algorithms
described in Chapter 6 of Klein, Driemel and Haverkort: Algorithmische 
Geometrie, 3rd edition, 2022. AGB-DTVD consists of seven files:
- agb-dtvd-main.cpp
- agb-dtvd-io.cpp (pre- and postprocessing operations)
- agb-dtvd-dc.cpp (this file: the Divide-and-Conquer algorithm)
- agb-dtvd-ricdd.cpp (randomised incremental construction with the 
    Delaunay DAG)
- agb-dtvd-ricpc.cpp (randomised incremental construction in rounds
    that follow a Polya curve)
- agb-simplepoint.cpp (the basis geometric operations)
- agb-qeds.cpp (the quad-edge data structure)


***********************************************************************/

#ifndef AGB_DTVD_DC_INCLUDED
#define AGB_DTVD_DC_INCLUDED


/**********************************************************************/

#include "agb-dtvd-io.cpp"

const char* dtvdAlgorithmDescription = 
  "divide and conquer";


/**********************************************************************/

// We are not going to store any vertex information, because, to keep
// exact computations simple, we are going to reduce all geometric 
// operations with vertices to an operation on four sites, which tests 
// whether the first site lies in the circumcircle of the other three. 
// Thus, we need to store only sites (= face information), no vertex 
// information. Later, however, to produce a drawing, we will have to 
// compute (approximate) vertex coordinates after all. Therefore we 
// still declare FPoint as the vertex type of a VoronoiEdge. Faces 
// of the Voronoi diagram correspond to Site objects, which have exact
// coordinates:

typedef HalfEdge<FPoint, Site> VoronoiEdge;

// in the dual, vertices and faces swap roles:

typedef HalfEdge<Site, FPoint> DelaunayEdge;


// This file provides two functions. The real work is done by 
// computeVDbyDC. It produces a Voronoi diagram that includes the
// following special vertices and edges:
// -  an "anchor" vertex that can be thought of as lying at (-infty, 0)
// -  for each unbounded end of a Voronoi edge, a vertex that acts as 
//    an endpoint that lies infinitely far in the direction of that 
//    edge;
// -  a cycle of edges that connect the vertices at infinity (including
//    the anchor).
//
// The return value of computeVDbyDC is an edge that starts at the 
// anchor, with the outer face to the right.
//
// The anchor enables two simplifications of the algorithm in the book: 
// 1. it enables us to represent a Voronoi diagram of one site as a 
//    correct QEDS, so that we need only one base case (one site) 
//    instead of two (two and three sites)
// 2. a starting point for the simultaneous traversal of the boundary 
//    cycles of two partial Voronoi diagrams is immediately given, we
//    do not need to search for such a starting point.

VoronoiEdge* computeVDbyDC(const Site sites[], int n);


// The next function is just a wrapper that removes the anchor vertex,
// so that postprocessing operations do not choke on it. The return 
// value is the edge on which the anchor used to be, with the outer 
// face to the right.

VoronoiEdge* computeVD(std::vector<Site> sites);

 
// Both functions assume that the sites are given in lexicographical 
// order.


/**********************************************************************/

VoronoiEdge* computeVDbyDC(const Site sites[], int n)
{
  static long int vertexId = 0;

  assert(n > 0);

  // if n=1, we create only the anchor and a loop, with the region of 
  // the only point on the inside and the region of a sentinel site
  // at "infinity" on the outside. As vertex information for the anchor
  // we simply take a NULL pointer:
  if (n == 1) 
    return VoronoiEdge::createLoop(new Site(sites[0]), NULL, &infinity);

  // n > 1: divide into two sets (left and right) and compute their VDs 
  // recursively.
  // We do not spend any time looking for a suitable split: we simply 
  // split by lexicographical order (that is, as if using symbolic 
  // perturbation). One of the reasons that this is safe, is that the 
  // algorithm has no concept of up and down and does not specifically 
  // need the y-monotonicity of the bisector between the two subsets; 
  // the algorithm will be fine if the bisector is monotonous in some 
  // direction, for example w.r.t. a slightly perturbed vertical axis.

  VoronoiEdge* lVD = computeVDbyDC(sites, n/2);
  VoronoiEdge* rVD = computeVDbyDC(sites + n/2, n - n/2);

  // now combine the two diagrams in three steps:
  // 1. find starting regions for the left-right bisector
  // 2. cut the diagrams along the left-right bisector
  // 3. discarded the parts that are cut off and combine the parts
  //    that are to be kept.

  // 1. find starting regions.
  // we begin our search in the regions that have an unbounded
  // intersection just below the negative x-axis, that is, we start
  // at the anchors.

  VoronoiEdge* lStart = lVD;
  Site* lSite = lStart->leftFace();

  VoronoiEdge* rStart = rVD;
  Site* rSite = rStart->leftFace();

  // The unbounded regions are exactly those whose vertices appear on 
  // the convex hull. So we are looking for a pair of sites from the 
  // left and the right diagram that are neighbours on the joint 
  // convex hull:
  while (true)
  {
    VoronoiEdge* lNext = lStart->scright();
    Site* lNextSite = lNext->leftFace();

    VoronoiEdge* rNext = rStart->scright();
    Site* rNextSite = rNext->leftFace();

    if (rightTurn(*lSite, *rSite, *rNextSite))
    {
      // rSite is not on the convex hull; move over:
      rSite = rNextSite;
      rStart = rNext;
      continue;
    }

    if (
      leftTurn(*lSite, *lNextSite, *rSite) ||
      collinearInOrder(*lSite, *lNextSite, *rSite)
    )
    {
      // lSite and cannot be a neighbour to rSite on the convex 
      // hull; move over:
      lSite = lNextSite;
      lStart = lNext;
      continue;
    }

    break;
  }

  // 2. cut along the left-right bisector

  // Now we need cut the two VDs along the left-right bisector. We will
  // maintain six half-edge pointers, three on each side:
  // lCut, rCut have their origin at the cutting point and point down,
  // with the cut to the left (when looking down);
  // lFace, rFace have their origin at the cutting point and the face
  // to be cut to the left (they will be the successors around the 
  // origin of lCut and rCut, respectively);
  // lExit, rExit are the edges that contain the projected exit points 
  // of these faces; lExit has that face to the left; rExit has it to 
  // the right.

  // first we create the starting point for the new bisector in both 
  // diagrams:

  lStart->insertVertex();
  VoronoiEdge* lCut = lStart->rev();
  VoronoiEdge* lFace = lCut->sccorgn();

  rStart->insertVertex();
  VoronoiEdge* rCut = rStart->rev();
  VoronoiEdge* rFace = rCut->sccorgn();

  // update lStart and rStart so that they will start at the starting 
  // point of the cut, following the boundary of the outer face on
  // the side that we will not be throwing away:

  lStart = lCut;
  rStart = rFace;

  // initialise lExit and rExit such that we start searching for the 
  // projected exit point on the first edge that is not part of 
  // the cut itself:

  VoronoiEdge* lExit = lFace; // search counterclockwise
  VoronoiEdge* rExit = rCut;  // search clockwise

  // for an example of the current state, see Figure 1 in the pdf;
  // the dashed segment indicates the initial part of the left-right
  // bisector, which is not in the QEDS yet at this point.

  do
  {
    // next we initialise a (new) segment of the cut from the current 
    // point (further) into the current face; this inserts (in each 
    // diagram) an edge, a vertex, and an edge that cut into that face; 
    // the location of the vertex is left unspecified (in fact, we leave 
    // the locations of vertices unspecified throughout the whole algorithm!)

    lCut = VoronoiEdge::cutIntoFace(lCut, lFace); lFace = lCut->sccorgn();
    rCut = VoronoiEdge::cutIntoFace(rCut, rFace); rFace = rCut->sccorgn();

    // see Figures 2 and 4 for an example of the updated values of
    // lCut, lFace, rCut and rFace

    // find prospective exit point in diagram for the sites on the left:
    // as long as the vertex and the end of lExit, which is equidistant to
    // lSite, lExit.rightFace(), and lExit.sccleft().rightFace(), would 
    // not exist in the diagram if we add rSite, we are still on the 
    // rSite-side of the left-right-bisector and we should continue our 
    // counterclockwise search along the boundary of lSite's region.
    // When lExit ends at the anchor, we should definitely stop (we 
    // check for this separately; otherwise there is trouble when there
    // is only one site).

    while (
      rSite->inOpenDisk(*lSite, *lExit->rightFace(), *lExit->sccleft()->rightFace())
      && lExit->sccleft() != lVD
    )  
      lExit = lExit->sccleft(); 

    // similarly in the diagram of the sites on the right. Now we should make
    // sure to stop before we are back at the starting point. On the other
    // hand, we should continue at the anchor. To achieve this, as the third 
    // region at the end of lExit we take lExit.sccleft().rightFace(), which 
    // is then equal to lexit.rightFace() and undefined; inOpenDisk 
    // returns true in that case.

    while (
      lSite->inOpenDisk(*rSite, *rExit->scright()->leftFace(), *rExit->leftFace())
      && rExit->rev() != rStart
    )
      rExit = rExit->scright(); 

    // see Figure 2 for an example of the updated values of lExit and rExit

    // now our prospective exit points are the circumcentres of
    // (lSite, rSite, lExit->rightFace()) and
    // (lSite, rSite, rExit->leftFace())
    // whichever of these does not contain the remaining point of these
    // four comes first, and we may extend the cut to that point:

    // decide how far we can extend the cut
    bool lExtendToExit =
      !rExit->leftFace()->inOpenDisk(*lSite, *rSite, *lExit->rightFace());
    bool rExtendToExit =
      !lExit->rightFace()->inOpenDisk(*lSite, *rSite, *rExit->leftFace());

    if (lExtendToExit)
    {
      // insert a vertex on the edge across which we enter the next
      // region (in degenerate cases it may happen that the crossing 
      // point coincides with an existing vertex; creating a new vertex 
      // for the crossing then results in a zero-length edge. This is on
      // purpose: we need that zero-length edge if we want the dual of
      // the Voronoi diagram to be a Delaunay triangulation):
      lExit = lExit->insertVertex();

      // move the current endpoint of the cut (the origin of lFace) 
      // onto the newly created vertex (the origin of lExit):
      VoronoiEdge::joinOrigins(lFace, lExit);

      // update lFace to be the edge from the newly created vertex
      // that has the face now entered on its left, and update lSite 
      // and lExit (where to continue the search in the next round)
      // accordingly:
      lFace = lFace->sccorgn();
      lSite = lFace->leftFace();
      lExit = lFace;

      // see Figure 3 for an example of the current state
    }

    // now before we proceed to the crossing on the right, note that
    // with degenerate inputs there is a catch: if the exits in 
    // the left-side diagram and the right-side diagram coincide (because 
    // the four relevant points are cocicular), we do not want to cross 
    // the boundaries on the left and on the right simultaneously. Rather
    // we cross on the left first, and save the crossing on the right for
    // the next round. Thus we will introduce a zero-length edge between
    // the two crossings, which will be necessary to ensure that the dual 
    // is a triangulation. Only when we cross into region of the point at 
    // infinity, we want to do so on both sides at once:

    if (lExtendToExit && !lSite->atInfinity()) rExtendToExit = false;

    // now we can handle the crossing on the right, if we still want it
    // (the code is slightly different from the code for the left side, 
    // because rExit has the face that we are leaving to the right,
    // whereas lExit had it to the left):

    if (rExtendToExit)
    {
      // careful here: rExit may be rStart.rev(), and, by the way the
      // QEDS works, inserting a vertex on rExit would erroneously 
      // change the starting point of rStart; to prevent this, we insert 
      // the new vertex from the other side, that is, on rExit.rev():
      rExit = rExit->rev()->insertVertex();
      VoronoiEdge::joinOrigins(rFace, rExit);
      rFace = rExit->pccorgn();
      rSite = rFace->leftFace();
    }
  } while (
    // As soon as we leave towards infinity in one of the diagrams, this
    // must also be the case in the other diagram, and we stop tracing the
    // bisector:
    !lSite->atInfinity()
  );

  // verify that nothing weird happened: the bisector indeed goes to 
  // infinity in the other diagram too:
  assert(rSite->atInfinity());

  // see Figure 5 for an example of the current state

  // 3. discard what is not needed and stitch the rest together:

  lCut->deleteSubgraph(lFace);
  rFace->deleteSubgraph(rCut); 

  // see Figure 6 for an example of the current state

  // careful now: rVD is now invalid; the anchor of the diagram of the 
  // subset to the right was in the part that we have just thrown away

  // now attach the two remaining parts to each other at the start and 
  // at the end of the left-right bisector:

  VoronoiEdge::insertGraph(lStart, rStart->pccorgn());
  VoronoiEdge::joinOrigins(lCut, rFace);

  // see Figure 7 for an example of the current state

  // then zip up the remaining gap along the left-right bisector:
  lCut->zipUp();

  // see Figure 8 for an example of the current state

  // now everything is in the same graph as the edge that originates 
  // from the anchor of the diagram of the left subset; this anchor
  // and this edge are still there and now represent the combined 
  // diagram:
  return lVD;
}


/**********************************************************************/

// computeVD is just a small wrapper function whose only purpose is to 
// remove the anchor vertex, so that postprocessing algorithms do not 
// choke on a meaningless vertex of degree two. 
// computeVD must be called with an array of sites in lexicographical 
// order.

VoronoiEdge* computeVD(std::vector<Site> sites)
{
  // verify that there are at least two sites. computeVDbyDC would work
  // correctly with just one site, but if there is only one site, we 
  // would get in trouble afterwards when trying to remove the anchor:
  assert(sites.size() > 1);

  // run the divide-and-conquer algorithm:
  VoronoiEdge* vd = computeVDbyDC(&*sites.begin(), sites.size());

  // remove the anchor vertex by zipping up the corresponding two-edge
  // face in the dual:
  vd->rtol()->zipUp();

  return vd; 
}


/**********************************************************************/

#endif

/***********************************************************************

RECENT CHANGES
0.02 Adapted to change of interface of the following QEDS operations:
     insertGraph, zipUp, cutIntoFace
0.03 Moved a few updates in and before the main while loop to make the 
     structure of the code easier to understand
0.05 Adapted to new interface of QEDS

***********************************************************************/

