/***********************************************************************

AGB-DTVD v0.06 BETA by Herman Haverkort, 10 May 2022

************************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of AGB-DTVD may be available from 
http://herman.haverkort.net.


************************************************************************

Was ist AGB-DTVD?

AGB-DTVD implements three algorithms to compute Voronoi diagrams and 
Delaunay triangulations. Its purpose is to illustrate the algorithms
described in Chapter 6 of Klein, Driemel and Haverkort: Algorithmische 
Geometrie, 3rd edition, 2022. AGB-DTVD consists of seven files:
- agb-dtvd-main.cpp
- agb-dtvd-io.cpp (pre- and postprocessing operations)
- agb-dtvd-dc.cpp (the Divide-and-Conquer algorithm)
- agb-dtvd-ricdd.cpp (this file: randomised incremental construction 
    with the Delaunay DAG)
- agb-dtvd-ricpc.cpp (randomised incremental construction in rounds
    that follow a Polya curve)
- agb-simplepoint.cpp (the basis geometric operations)
- agb-qeds.cpp (the quad-edge data structure)


***********************************************************************/

#ifndef AGB_DTVD_RICDD_INCLUDED
#define AGB_DTVD_RICDD_INCLUDED

/**********************************************************************/

#include "agb-dtvd-io.cpp"

const char* dtvdAlgorithmDescription = 
  "randomised incremental construction with Delaunay DAG";


// This file provide a class DelaunayDAGNode from which we can build
// the Delaunay DAG:

class DelaunayDAGNode;


// For triangles in the Delaunay triangulation (and thus, the vertices
// in the Voronoi diagram) we will store pointers to nodes in the 
// Delaunay DAG; this is so that, when we remove a triangle from the
// triangulation, we can easily find the corresponding DAG node so that
// we can update it:
  
typedef HalfEdge<Site, DelaunayDAGNode> DelaunayEdge;
typedef HalfEdge<DelaunayDAGNode, Site> VoronoiEdge;


// Apart from the DelaunayDAGNode class, this file provides two 
// functions. To run the algorithm, call computeVD. This will then 
// take care of degeneracies and permute the input randomly; after
// that, computeVDbyIC does the real work. The return value is an
// (arbitrary) halfedge of the Voronoi diagram.

VoronoiEdge* computeVD(std::vector<Site> sites);
VoronoiEdge* computeVDbyIC(const Site sites[], int n);


/**********************************************************************/

// THE DELAUNAY DAG

// A DelaunayDAGNode is also an FPoint, so that it can later store the
// coordinates of the corresponding Voronoi vertex

class DelaunayDAGNode: public FPoint
{
  // A node in a Delaunay DAG represents a triangle of three sites, 
  // may have children, and, if it still exists in the current
  // triangulation, stores a pointer to a halfedge in the QEDS that
  // has the triangle on the left
  Site a; 
  Site b; 
  Site c;
  std::vector<DelaunayDAGNode*> children;
  DelaunayEdge* edge;

  // annulus that contains the circumcircle. This is used to speed up
  // queries. Only if a point lies between the inner and the outer
  // circle of the annulus, an exact point-in-circumcircle test is done:
  Site centre;
  Site::Number sqInnerRadius; 
  Site::Number sqOuterRadius; 

    public:
  // construct a DAG node for a triangle, specified by giving
  // a half-edge that has the triangle on the left:
  DelaunayDAGNode(DelaunayEdge* e = NULL):
    edge(e),
    a(*e->origin()), b(*e->destin()), c(*e->pccleft()->origin())
  {
    // compute an annulus that approximates the circumcircle:
    computeApproximateCircumcircle(a, b, c, centre, sqInnerRadius, sqOuterRadius);
  };

  // when a triangle is removed from the current triangulation, we
  // will use markRetired to make the halfedge pointer invalid; 
  // this will tell the point location algorithm that this triangle
  // cannot be returned as an answer anymore, and search should 
  // always continue in the children (if there are any):
  void markRetired() {
    // make the edge pointer invalid:
    edge = NULL;
  }

  void addChild(DelaunayDAGNode* n) {
    children.push_back(n);
  }

  // findTriangleContaining searches the DAG and returns an halfedge
  // that is an edge of a triangle in the current triangulation that 
  // contains the given point p. The answer is not uniquely defined:
  // as explained in Section 6.2.1. of the book, a point can be 
  // contained in multiple unbounded triangles; a point could also 
  // lie on an edge between bounded triangles. In such cases, 
  // findTriangleContaining returns just one triangle that contains 
  // the point; the RIC algorithm will find the others via edge flips. 
  DelaunayEdge* findTriangleContaining(const Site& p) {
    Site::Number sqR = sqDistance(p, centre);
    if (sqR > sqOuterRadius) return NULL;
    if (sqR >= sqInnerRadius && 
      // if sqR lies between sqInnerRadius and sqOuterRadius, we do
      // an exact test:
      !p.inOpenDisk(a, b, c)
    ) 
      // return invalid edge (point not found in this branch):
      return NULL; 

    if (edge != NULL && p.inClosedTriangle(a, b, c))
      return edge;

    // otherwise check the children:
    for (int i = 0; i < children.size(); ++i)
    {
      DelaunayEdge* e = children[i]->findTriangleContaining(p);
      if (e) return e;
    }

    // return invalid edge (point not found in this branch):
    return NULL;
  }

  // A DelaunayDAGNode also has the x and y attributes of FPoint, 
  // so that we can store the coordinates of the corresponding 
  // Voronoi vertex in it. This is for postprocessing. The 
  // randomised incremental construction algortihm will not use 
  // this.
  DelaunayDAGNode& operator=(const FPoint& p) {
    x = p.x; y = p.y; return *this;
  }
};


/**********************************************************************/

// THE INCREMENTAL CONSTRUCTION ALGORITHM:

VoronoiEdge* computeVDbyIC(const Site sites[], int n)
{
  // The incremental construction starts with a graph that consists of 
  // two faces: a triangle with two actual sites and a point at 
  // "infinity", and the complement of this triangle. These triangles 
  // represent two unbounded regions, one to the left of the line 
  // through the initial two sites, and one to the right. The line 
  // segment between the two sites is included in both regions. 

  // The point at infinity may be easiest to understand if we imagine
  // the Delaunay triangulation to be constructed in a small area on the 
  // front of a large sphere, with the point at "infinity" at the back. 
  // At all times, we will have a triangulation of actual sites on the 
  // front, with, for each site on their convex hull, an edge that
  // goes around to the point on the back. We can think of the initial 
  // configuration as three edges that together form a great circle, 
  // dividing the sphere in a left half and a right half.
 
  // To set up this initial configuration, we create a loop with 
  // one vertex, representing infinity, and a root node for the DAG:

  DelaunayEdge* dt = DelaunayEdge::createLoop(NULL, &infinity);
  DelaunayDAGNode* dag = new DelaunayDAGNode(dt);

  // then we make a copy of the first two sites and insert them 
  // in the loop:

  dt->insertVertex(new Site(sites[0]));
  dt->insertVertex(new Site(sites[1]));

  // for the infinite "triangle" on the left side, we create a DAG node, 
  // which we make a child of the root:

  DelaunayDAGNode* v;
  v = new DelaunayDAGNode(dt); dag->addChild(v); dt->labelLeftFace(v);
  
  // we do the same for the triangle on the other side:

  dt = dt->rev();
  v = new DelaunayDAGNode(dt); dag->addChild(v); dt->labelLeftFace(v);

  // we mark the root as retired, so that search will always be
  // directed to the children:

  dag->markRetired();

  // next we insert the remaining sites one by one:

  for (int i = 2; i < n; ++i)  
  {
    // make a copy of the site to insert:
    Site* p = new Site(sites[i]);

    // find a triangle that contains the site:
    DelaunayEdge* e = dag->findTriangleContaining(*p);
    assert(e);

    // this triangle is going to be removed from the triangulation, so 
    // mark it as retired in the DAG, so that search must always 
    // continue to the children:
    e->leftFace()->markRetired();

    // insert the new site into the triangle, with edges to the three
    // corners:
    e = e->insertStar(p);

    // initialise a stack with edges that need to be checked if they
    // need to be flipped; we will maintain that the edges on the stack
    // have the new site p to the left.
    std::stack<DelaunayEdge*> toTest;
    toTest.push(e->pccorgn()->sccleft());
    toTest.push(e->sccleft());
    toTest.push(e->sccorgn()->sccleft());

    // repeat until the retriangulated star around the new site is final:
    while (!toTest.empty())
    {
      // get an edge from the stack, and check if the new site p on the 
      // left violates the empty-circle property of the triangle to the
      // right:
      DelaunayEdge* e = toTest.top(); toTest.pop();
      assert(e->rightFace());
      if (p->inOpenDisk(*e->origin(), *e->pcright()->origin(), *e->destin()))
      {
        // if yes, then e must be flipped, which necessitates the 
        // following operations.
        // 1. the triangle R on the right is going to disappear from the 
        //    triangulation, so it has to be marked retired in the DAG:
        e->rightFace()->markRetired();

        // 2. the faces on each side of the edge must be labelled with
        //    the same DAG node, otherwise the QEDS may refuse the flip.
        //    we keep the label of the triangle R to the right, because,
        //    when all the flipping is done, the other two edges of R may 
        //    end up forming new triangles together with p, and then we 
        //    need to know R so that we can register the new triangles 
        //    as its children: 
        e->labelLeftFace(e->rightFace());

        // 3. Now we can safely do the flip:
        e->flip();

        // 4. And we put the other two edges of R on the stack to see if
        //    the triangles behind them are invalidated too.
        toTest.push(e->pccleft());
        toTest.push(e->pccorgn());
      }
      else
      {
        // the triangle to the right is still valid, so the edge e is 
        // on the boundary of star that had to be retriangulated. This 
        // means that e, together with p, forms a new triangle that 
        // needs to be inserted in the DAG, as a child of the 
        // triangles that were on both sides of e before:

        DelaunayDAGNode* newTriangle = new DelaunayDAGNode(e);
        e->leftFace()->addChild(newTriangle);
        e->rightFace()->addChild(newTriangle);

        // finally we label the new triangle in the triangulation 
        // with a pointer to its occurrence in the DAG:

        e->labelLeftFace(newTriangle);
      }
    }
  }

  // return an edge of the dual, that is, the Voronoi diagram:
  return dt->rtol();
}


/**********************************************************************/

// SETTING UP THE RANDOMISED INCREMENTAL CONSTRUCTION:

// The following function computes the Voronoi diagram for a given set
// of sites; these must be given in lexicographic order:

VoronoiEdge* computeVD(std::vector<Site> sites)
{
  // This is just a wrapper function. Most of it would not be necessary
  // if we could be sure that the input is in general position, but how
  // would we know? We better make sure that we do not choke on 
  // degeneracies.

  // Because of the way the RIC starts, we have to be a bit careful:
  // * there must at least two sites;
  // * the third site cannot be collinear with the initial two sites
  //   (this would initiate conflicting edge flips)
  // * any later sites that are collinear with the initial two must lie
  //   between the initial two sites (collinear sites that do not lie 
  //   between the initial two sites cannot be located in the Delaunay 
  //   DAG, because they lie outside both initial "triangles":

  assert(sites.size() >= 2);

  if (sites.size() > 2)
  {
    // make sure the sites[0] and sites[1] are the extremes on the 
    // line that contains them; assuming the input is given in 
    // lexicographic order, the following line does the trick:
    swap(sites[1], sites.back());

    // make sure sites[2] is not on that line:
    for (int i = 2; i < sites.size(); ++i)
      if (!collinear(sites[0], sites[1], sites[i]))
      {
        swap(sites[2], sites[i]);
        break;
      }
    assert(!collinear(sites[0], sites[1], sites[2]));
  }

  // permute the remaining sites randomly. NOTE: if we do not seed the 
  // random number generator, the order will be the same every time!
  permuteRandomly(&sites[3], sites.size()-3);

  // now we are ready to run the incremental construction algorithm:
  return computeVDbyIC(&sites[0], sites.size());
}


/**********************************************************************/

#endif


/***********************************************************************

RECENT CHANGES:
0.03 Small changes in the initialisation that should not have any real
     effect (other than that they may change the edge that is returned
     in the end, and thus, the order in which edges are presented to 
     the colouring algorithm etc.)
0.05 Adapted to new interface of QEDS
0.06 Added fast approximate point-in-circle tests to the DAG

***********************************************************************/

