/***********************************************************************

AGB-DTVD v0.06 BETA by Herman Haverkort, 10 May 2022

************************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of AGB-DTVD may be available from 
http://herman.haverkort.net.


************************************************************************

Was ist AGB-DTVD?

AGB-DTVD implements three algorithms to compute Voronoi diagrams and 
Delaunay triangulations. Its purpose is to illustrate the algorithms
described in Chapter 6 of Klein, Driemel and Haverkort: Algorithmische 
Geometrie, 3rd edition, 2022. AGB-DTVD consists of seven files:
- agb-dtvd-main.cpp
- agb-dtvd-io.cpp (pre- and postprocessing operations)
- agb-dtvd-dc.cpp (the Divide-and-Conquer algorithm)
- agb-dtvd-ricdd.cpp (randomised incremental construction with the 
    Delaunay DAG)
- agb-dtvd-ricpc.cpp (this file: randomised incremental construction in 
    rounds that follow a Polya curve)
- agb-simplepoint.cpp (the basis geometric operations)
- agb-qeds.cpp (the quad-edge data structure)


***********************************************************************/

#ifndef AGB_DTVD_RICPC_INCLUDED
#define AGB_DTVD_RICPC_INCLUDED

/**********************************************************************/

#include "agb-dtvd-io.cpp"

const char* dtvdAlgorithmDescription = 
  "incremental construction in randomised rounds, each following a plane-filling curve, with point location by walking";

// The triangulation algorithm does not compute coordinates of Voronoi
// vertices. However, to be able to fill them in later, we declare F
// Point as the vertex type of a VoronoiEdge, and as the face type of a
// DelaunayEdge. Faces of the Voronoi diagram / vertices of the 
// Delaunay triangulation correspond to Site objects, which have exact
// coordinates:

typedef HalfEdge<FPoint, Site> VoronoiEdge;
typedef HalfEdge<Site, FPoint> DelaunayEdge;

// To run the algorithm, call computeVD. This will then 
// take care of degeneracies and permute the input; after
// that, computeVDbyIC does the real work. The return value is an
// (arbitrary) halfedge of the Voronoi diagram.

VoronoiEdge* computeVD(std::vector<Site> sites);
VoronoiEdge* computeVDbyIC(const Site sites[], int n);


/**********************************************************************/

// THE INCREMENTAL CONSTRUCTION ALGORITHM:

VoronoiEdge* computeVDbyIC(const Site sites[], int n)
{
  // The incremental construction starts with a graph that consists of 
  // two faces: a triangle with two actual sites and a point at 
  // "infinity", and the complement of this triangle. These triangles 
  // represent two unbounded regions, one to the left of the line 
  // through the initial two sites, and one to the right. The line 
  // segment between the two sites is included in both regions. 

  // The point at infinity may be easiest to understand if we imagine
  // the Delaunay triangulation to be constructed in a small area on the 
  // front of a large sphere, with the point at "infinity" at the back. 
  // At all times, we will have a triangulation of actual sites on the 
  // front, with, for each site on their convex hull, an edge that
  // goes around to the point on the back. We can think of the initial 
  // configuration as three edges that together form a great circle, 
  // dividing the sphere in a left half and a right half.
 
  // To set up this initial configuration, we create a loop with 
  // one vertex, representing infinity:

  DelaunayEdge* dt = DelaunayEdge::createLoop(NULL, &infinity);

  // then we make a copy of the first two sites and insert them 
  // in the loop:

  dt->insertVertex(new Site(sites[0]));
  dt->insertVertex(new Site(sites[1]));

  // next we insert the remaining sites one by one:

  for (int i = 2; i < n; ++i)  
  {
    // make a copy of the site to insert:
    Site* p = new Site(sites[i]);

    // find a triangle that contains the site, starting from the
    // current edge (dt) that is incident on the previous point:
    for (int edgesChecked = 0; edgesChecked < 3; )
    {
      // edgesChecked is the number of edges of the current triangle,
      // for which we have verified that we are on the correct side
      // of the line that contains the edge
      if ( 
        dt->origin()->atInfinity() || 
        dt->destin()->atInfinity() || 
        !rightTurn(*dt->origin(), *dt->destin(), *p)
      )
      { // we are on the correct side, stay in this triangle and 
        // go to the next edge to check:
        ++edgesChecked;
        dt = dt->sccleft();
      }
      else 
      { // we are on the wrong side of this edge, go the neighbour
        // on the other side and check the next edge:
        dt = dt->pccorgn();
        edgesChecked = 1;
      }
    }

    // insert the new site into the triangle, with edges to the three
    // corners:
    dt = dt->insertStar(p);

    // initialise a stack with edges that need to be checked if they
    // need to be flipped; we will maintain that the edges on the stack
    // have the new site p to the left.
    std::stack<DelaunayEdge*> toTest;
    toTest.push(dt->pccorgn()->sccleft());
    toTest.push(dt->sccleft());
    toTest.push(dt->sccorgn()->sccleft());

    // repeat until the retriangulated star around the new site is final:
    while (!toTest.empty())
    {
      // get an edge from the stack, and check if the new site p on the 
      // left violates the empty-circle property of the triangle to the
      // right:
      DelaunayEdge* e = toTest.top(); toTest.pop();
      if (!p->inOpenDisk(*e->origin(), *e->pcright()->origin(), *e->destin()))

        // if not, there is nothing to do for this edge, check the next:
        continue;

      // otherwise, e must be flipped:
      e->flip();

      // and we put the other two edges of R on the stack to see if
      // the triangles behind them are invalidated too:
      toTest.push(e->pccleft());
      toTest.push(e->pccorgn());
    }
  }

  // return an edge of the dual, that is, the Voronoi diagram:
  return dt->rtol();
}


/**********************************************************************/

// SETTING UP THE INCREMENTAL CONSTRUCTION:

// The following function computes the Voronoi diagram for a given set
// of sites; these must be given in lexicographic order:

VoronoiEdge* computeVD(std::vector<Site> sites)
{
  // This is a wrapper function. Some of it would not be necessary
  // if we could be sure that the input is in general position, but how
  // would we know? We better make sure that we do not choke on 
  // degeneracies.

  // Because of the way the RIC starts, we have to be a bit careful:
  // * there must at least two sites;
  // * the third site cannot be collinear with the initial two sites
  //   (this would initiate conflicting edge flips)
  // * any later sites that are collinear with the initial two must lie
  //   between the initial two sites (collinear sites that do not lie 
  //   between the initial two sites cannot be located in the Delaunay 
  //   DAG, because they lie outside both initial "triangles":

  assert(sites.size() >= 2);

  if (sites.size() > 2)
  {
    // make sure the sites[0] and sites[1] are the extremes on the 
    // line that contains them; assuming the input is given in 
    // lexicographic order, the following line does the trick:
    swap(sites[1], sites.back());

    // make sure sites[2] is not on that line:
    for (int i = 2; i < sites.size(); ++i)
      if (!collinear(sites[0], sites[1], sites[i]))
      {
        swap(sites[2], sites[i]);
        break;
      }
    assert(!collinear(sites[0], sites[1], sites[2]));
  }

  // permute the remaining sites randomly. NOTE: if we do not seed the 
  // random number generator, the order will be the same every time!
  clock_t started = clock();
  permuteRandomly(&sites[3], sites.size()-3);
  cerr << "(permuting: " << 1000.0*(clock()-started)/CLOCKS_PER_SEC << "ms)...";
  cerr.flush();

  // for rounds of tripling sizes, sort the remaining sites along the
  // Polya curve; the last round is sorted first, it consists of the last 2/3 
  // of the sequence; we do not sort rounds before sites[9], because sites[0..2]
  // should stay where they are in the order. 
  started = clock();
  bool reverse = false;
  for (int i = sites.size(); i > 8; i /= 3)
  {
    sort(sites.begin() + i/3, sites.begin() + i, Site::PolyaOrder());
    if (reverse) // actually reversing every other round hardly has any effect
    {
      int j = i/3; int k = i-1;
      for (; j < k; ++j, --k) 
        swap (sites[j], sites[k]);
    }
    reverse = !reverse;
  }
  cerr << "(sorting: " << 1000.0*(clock()-started)/CLOCKS_PER_SEC << "ms)...";
  cerr.flush();

  // now we are ready to run the incremental construction algorithm:
  return computeVDbyIC(&sites[0], sites.size());
}


/**********************************************************************/

#endif


/***********************************************************************
***********************************************************************/

